/** ----------------------------------- -o-
 *  Automata.java
 *
 * Provide an array of threads scheduled by independent timers.
 * Allow user to define, start and stop each thread singly and as
 * a group.
 *
 * Threads are defined by a timer-thread algorithm and passed a
 * string of arguments.  User chooses which index to assign the
 * thread.  The size of the list of indices is given as an
 * argument to Automata.
 *
 * All threads are stopped until started.  start creates a new
 * if the user has sent a timer-thread definition.  otherwise it
 * restarts the existing thread if one exists.  restart stops then starts.  
 * Posting a definition of a timer-thread algorithm does not automatically 
 * start or restart an existing thread.
 *
 *
 * INLETS--
 *   0: (message)  directives
 * OUTLETS--
 *   N: (list)  one for each thread
 *
 * MESSAGES--
 *   stop <n> / stopall
 *   start <n> / startall 
 *   restart <n> / restartall 
 *   thread <n> <algo> <args...>
 *
 * CLASSES/METHODS--
 *   Child() 	simple structure for timer-thread and support
 *
 * @version 0.1    CREATED: Thu Sep 14 00:00:00 PDT 2009
 * @author  dreeder@danaprajna.com
 */ 							// -o-
package com.danaprajna;

import java.lang.Runtime;
import java.util.Vector;

import com.cycling74.max.*; 		


public class Automata extends MaxObject 
{ 							
  //----------------------------------- --o-
  // variables 			
  //
  private DebugLabel 	z      = new DebugLabel("Automata", true);
  private String 	USAGE  = "Usage: Automata() <num_threads>";

  private ToolBox t = new ToolBox();

  private Child[] children = null;



  //----------------------------------- --o-
  // constructors 					
  //
  private class Child { 
    TimerThread  c  = null; 	// child
    Thread 	 t  = null;  	// thread bearing child
    Atom[] 	 a  = null; 	// childish arguments
      				//   XXX  a[0]: name of timer-process algorithm
  }

  public Automata(Atom[] args) 	
  {
    // z.b(args);
    if ( (1!=args.length) || (!args[0].isInt()) )  
      { z.merr("<args> is the wrong length or wrong type"); return; }

    children = new Child[args[0].getInt()];
    for (int i=0; i<children.length; i++)  { children[i] = new Child(); }
      

    declareInlets(new int[] {DataTypes.MESSAGE} );
    setInletAssist(0, "(message)  define and manage timer-processes");

    int[] outletArray = new int[children.length];
    for (int i=0; i<children.length; i++) {
      outletArray[i] = DataTypes.LIST;
      setOutletAssist(i, "(list)  data from timer-process #" + i);
    }
    declareOutlets(outletArray);
    createInfoOutlet(false);

    // z.e();
  }  // public Automata(Atom[])


  protected void notifyDeleted()  
    { stopall(); z.sm("notifyDeleted").d("all threads stopped").ms(); }



  //----------------------------------- --o-
  // built-in messages 					
  //
  public void usage() { usage(true); }
  public void usage(boolean do_longversion) 
  {
    String space = "\t\t\t\t\t";
    z.pc(USAGE); if (false == do_longversion) { return; } 

    z.pc(space + "thread <idx> <algorithm> <args...>");
    z.pc(space + "(re)start <idx>");
    z.pc(space + "(re)startall");
    z.pc(space + "stop <idx>");
    z.pc(space + "stopall");

    for (int i=0; i<children.length; i++)
      { if (null != children[i].c)  { children[i].c.usage(); break; } }
      // XXX  also serves as indicator of child instantiation
  }

  public void anything(String m, Atom[] args) 
    { z.arglist("anything", m, args); }



  //----------------------------------- --o-
  // messages 						
  //
  public void out(int id, String s) { 
      if (true != outlet(id, s))  
        { z.sm("out").merr("failed outlet(" + id + ")").ms(); } 
    }  
  public void out(int id, double d)  { out(id, d + ""); }

  
  public void stopall() { 
      //z.sm("stopall").d().ms(); 
      for (int i=0; i<children.length; i++)  { stop(i); } 
    }
  public void stop(int i)  
    { if (null!=children[i].c)  { 
	//z.sm("stop").d("(idx=" + i + ")").ms(); 
	children[i].c.stop();
      } }


  public void startall() {   
      //z.sm("startall").d().ms(); 
      for (int i=0; i<children.length; i++)  { start(i); }
    }
  public void start(int i) 
    { if (null!=children[i].c) { 
	//z.sm("start").d("(idx=" + i + ")").ms();
	children[i].c.start(); 
      } } 


  public void restartall() { 
      //z.sm("restartall").d().ms(); 
      for (int i=0; i<children.length; i++)  { restart(i); } 
    }
  public void restart(int i)  
    { if (null!=children[i].c) { 
        //z.sm("restart").d("(idx=" + i + ")").ms(); 
        stop(i); start(i); 
    } }


  // "thread" <idx> <object_type> <args...> 		// -o-
  //
  // Handle directive from patcher environment: choose an index and
  //   pass shortened argument list.  Instantiation happens in start().
  //
  // <idx> 		[0-count, children.length)
  // <object_type> 	handled in start()
  // <args...> 		handled by constructor for <object_type>
  //
  public void thread(Atom[] args) { 
    z.sm("thread");	// z.b("thread", args);
    String lusage = "thread <idx> <object_type> <args...>";

    if ( (3>args.length) || !(args[0].isInt() && args[1].isString()) )
      			{ z.merr(lusage).ms(); 			return; }
    int idx = args[0].getInt();
    if ( (0>idx) || (children.length<=idx) ) 	
      			{ z.merr("<idx> is out of range").ms(); return; }


    // create a thread for this child
    Child lc = children[idx];

    lc.a = Atom.removeFirst( args );
    lc.c = new TimerThread(this, idx, lc.a); 
    lc.t = new Thread(lc.c); 		// NB!  no need for lc.t.start()

    // z.e();
  }  // thread()



  //----------------------------------- --o-
  // something risque 			
  // 
  public void experiment(Atom[] args) { 
    z.arglist("experiment", args);  // z.b("experiment");
    try {

      if (false) { 			// TEST: fib
        String s = "";
        for (int i=1; i<20; i++)  { s += t.fib(i) + " "; }
	z.pr(s);
      }

      if (false) {  			// TEST: Urn
        int size = 7;
        Urn u = new Urn(size);

	int count = 0;
	int max = 24;
	String list = "";

	while (count++<max) {
	  list += u.next() + " ";
	  if (0==(count%size))  { z.pr(list); list = ""; }
	}

	z.pr(list);


      } // TEST: urn

      if (false) { 			// TEST: start/restart/stop
        startall(); stopall(); }

    } 
    catch (Exception e)  
      { bail(e.getMessage()); e.printStackTrace(); }


    for (Child c : children)  { if (null!=c.c)  { c.c.experiment(); } }
     
    //z.e();
  }  // experiment()

  public void experiment()  { Atom[] a = null; experiment(a); }

}  // public class Automata extends MaxObject
  
