/** ----------------------------------- -o-
 *  TimerThread.java
 *
 * container for timer-process algorithm class.  
 * iterates algorithm after each MaxClock delay. 
 * 
 * INPUT--
 *   parameters to define algorithm context
 * OUTPUT--
 *   per each timer iteration: output to indicated outlet (posted by parent)
 * MESSAGES--
 *   <constructor> 	defines initial context
 *   start 		start clock
 *   stop 		stop clock
 * 
 * TBD  add new timer-process algorithm classes as neighbors to existing
 * 	  algorithms
 *
 * XXX  optimize by coercing all algorithm classes to MaxObject?
 * TBD FIX  what if one thread is initialized with more than one algo?
 *
 * @version 0.1    CREATED: Thu Sep 14 00:00:00 PDT 2009
 * @author  dreeder@danaprajna.com
 */  							// -o-
package com.danaprajna;

import java.util.Vector;
import com.cycling74.max.*; 		


public class TimerThread extends MaxObject implements Runnable
{ 							
  //----------------------------------- --o-
  // static variables 			
  //
  private DebugLabel 	z      = new DebugLabel("TimerThread", true);
  private String  	USAGE  = "Usage: TimerThread() "
    + "<parent> <idx> <args...>";

  private Automata  parent    = null;  	// output channel to patch environment
  private int 	    threadId  = -1;
  private String    algoName  = null;
  private Atom[]    args      = null;



  //
  // dynamic variables 		
  //
  private boolean   running  = false; 
  private MaxClock  mc 	     = null; 	
    // defined only when timeloop() running
    // TBD  why not use java.lang.Object.wait()?

  private StepSwing ss = null;



  //----------------------------------- --o-
  // constructors 					
  //
  public TimerThread(Automata p, int idx, Atom[] a)  
  {
    z.st(idx);	// b(a);

    if (null==p)      { z.merr("parent is undefined"); 		return; }
    if (0>idx)        { z.merr("<idx> is less than zero"); 	return; }
    if (1>a.length)
      		      { z.merr("<args> length is too short"); 	return; }
    if (! a[0].isString()) { 
    			z.merr("first of <args> is not an algorithm name"); 
								return; }

    parent    = p;
    threadId  = idx;
    algoName  = a[0].toString();
    args      = Atom.removeFirst(a);


    // build algorithm object
    //
    try {
      if (algoName.equals("StepSwing")) {
	if (null==ss)  { ss = new StepSwing(parent, threadId, args); }; }
      else 
	{ z.merr("Unknown algorithm: " + algoName); 		return; } 

    }
    catch (Exception e)
      { bail(e.getMessage()); e.printStackTrace(); }

    // z.e();
  }  // public TimerThread(Atom[])


  protected void notifyDeleted()  
    { z.sm("notifyDeleted").d("stop").ms(); stop(); }
  
  public void run()  { /*EMPTY*/ }



  //----------------------------------- --o-
  // built-in messages 					
  //
  public void usage() { usage(true); }
  public void usage(boolean do_longversion) {  // XXX  no short version
    if (false==do_longversion) { return; } 
    if (null==algoName)  { return; }

    if (algoName.equals("StepSwing"))  { ss.usage(); }
  }



  //----------------------------------- --o-
  // messages 						
  //
  public TimerThread stop() { 
      // z.b("stop");

      if (null != mc)  { mc.unset(); mc.release(); mc = null; }
      running = false; 

      if (algoName.equals("StepSwing"))  { ss.stop(); }

      // z.e();
      return(this); 
    }

  public TimerThread start() {
      // z.b("start");

      if (running)  { return(this); } 	// already started
      running = true;
      if (null == mc)  { mc = new MaxClock(this, "timeloop"); }

      if (algoName.equals("StepSwing"))  { ss.start(); }

      // z.e("before delay");
      mc.delay(0.0);
      return(this); 	// XXX  never reached?
    }

  public void timeloop() 
    { if (!running) { stop(); return; };  mc.delay(runalgo()); } 


  // runalgo 	
  //
  // Invoke indicated timer-process.
  // RETURN: algorithm returns delay time in <ms> 
  //
  public double runalgo() { 
    // z.b("runalgo");
    double delay = 0.0;

    try {
      if (algoName.equals("StepSwing"))  { delay = ss.algorithm(); }

    }
    catch (Exception e)  
      { bail(e.getMessage()); e.printStackTrace(); }

    // z.e();
    return(delay); 
  }  // runalgo()

  

  //----------------------------------- --o-
  // something risque 			
  // 
  public void experiment(Atom[] args) { 
    z.arglist("experiment", args);  //z.b("experiment");
    
    try {
      if (algoName.equals("StepSwing"))  { ss.experiment(); }

    } 
    catch (Exception e)
      { bail(e.getMessage()); e.printStackTrace(); }

    //z.e();
  }  // experiment()

  public void experiment()  { Atom[] a = null; experiment(a); }

}  // public class TimerThread extends MaxObject

